// ========================================================================================== //
//             ___  _   _    __    __  __    __    ____  __    ____  _____  _  _              //
//            / __)( )_( )  /__\  (  \/  )  /__\  ( ___)(  )  ( ___)(  _  )( \( )             //
//           ( (__  ) _ (  /(__)\  )    (  /(__)\  )__)  )(__  )__)  )(_)(  )  (              //
//            \___)(_) (_)(__)(__)(_/\/\_)(__)(__)(____)(____)(____)(_____)(_)\_)             //
//            -------------------------------------------- G A M E   E N G I N E              //
//                                                                                            //
// ========================================================================================== //
// Developed and implemented by:                                                              //
//     Oskam Thomas                                                                           //
//                                                                                            //
//                                                                                            //
// Copyright 2007/2008. All rights reserved.                                                  //
// ========================================================================================== //





// ============================================================================================
// ---- Shader Parameter ----------------------------------------------------------------------
// ============================================================================================


// view projection matrix
float4x4 xViewProjection;


// translation in world coordinates that is applied to the whole heightfield
float3 xHeightfieldTranslation;


// factor with which the heightmap height values are multiplied, in world coordinates
float xMaxAlteringHeight;


// factor multiplied to the index (x,y number) of each instanced patch, to get the texture 
// coords on the heightmap 
float2 xHeightfieldPixelOffset;


// world size of a patch (patch is quadratic)
float xPatchWidth;





// ============================================================================================
// ---- Internal VertexSh-PixelSh parameter definitions ---------------------------------------
// ============================================================================================




// normal Vertex Shader Input
struct VS_input {
	float4 Position		: POSITION;
	float2 TexCoords	: TEXCOORD0;
};



// normal Pixel Shader input
struct Fast_Z_PS_input {
    float4 Position			: POSITION; 
};



// normal Frame Buffer input
struct FB_input {
	float4 Color		: COLOR0;
};





// GLOBAL TEXTURES: each pixel defines a value for the grid
//=============================================================================================

// heightfield texture: used to perturbate the height of the grid
uniform Texture xHeightfield;
sampler HeightTextureSampler = sampler_state
{
	texture = <xHeightfield>;
	//magfilter = LINEAR;
	//minfilter = LINEAR;
	//mipfilter = LINEAR;
	magfilter = NONE;
	minfilter = NONE;
	mipfilter = NONE;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};







// ============================================================================================
// ---- Vertex Shader  ------------------------------------------------------------------------
// ============================================================================================




// Vertex shader for a fast depth buffer pass (needed for proper alpha blending of patches)
Fast_Z_PS_input Fast_Z_Pass_VS( VS_input input, float2 PatchIndex : TEXCOORD2 )
{
	// 0) declare Output and initialize all variables to 0
	Fast_Z_PS_input Output = (Fast_Z_PS_input)0;
	
	// 1) get heightfield texture value & compute height value
	float2 globalTexCoords = input.TexCoords + PatchIndex * xHeightfieldPixelOffset;
	float texColor = tex2Dlod(HeightTextureSampler, float4(globalTexCoords,0.0f,1.0f)).x;
	float3 h = {0.0f, xMaxAlteringHeight * texColor, 0.0f};
	
	// 3) compute translation
	float3 t = {PatchIndex.x * xPatchWidth, 0.0f, PatchIndex.y * xPatchWidth};

	
	// 4) store output
	float4 finalPos4 = float4(input.Position + h + t + xHeightfieldTranslation, 1.0f);
	Output.Position = mul(finalPos4, xViewProjection);
	return Output;
}











//------------------------------------------------------------------------------------------
// Helper Function: computes mipmap level --------------------------------------------------
//------------------------------------------------------------------------------------------

float MipmapLevel(float2 uv, float2 textureSize)
{
	float2 dx = ddx(uv * textureSize.x);
	float2 dy = ddy(uv * textureSize.y);
	float d = max( dot(dx, dx), dot(dy, dy) );

	//return log2( sqrt(d) );
	return 0.5 * log2( d );
}









// ============================================================================================
// ---- Vertex Shader  ------------------------------------------------------------------------
// ============================================================================================



// Pixel Shader for a fast z pass
FB_input Fast_Z_Pass_PS(Fast_Z_PS_input input)
{
	FB_input Output = (FB_input)0;
	Output.Color = float4(1,1,1,0);
	return Output;
}






// ============================================================================================
// ---- Shader Techniques  --------------------------------------------------------------------
// ============================================================================================

technique fastZPass
{
pass
    {
        VertexShader = compile vs_3_0 Fast_Z_Pass_VS();
        PixelShader = compile ps_3_0 Fast_Z_Pass_PS();
    }
}