using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using System.Text;

namespace HFShader.Tools
{

    public struct IRay
    {
        // fields
        public Vector3 Position;
        public Vector3 EndPosition;
        public Vector3 Direction;
        public float Max_t;

        // constructor
        public IRay(Vector3 start, Vector3 end)
        {
            Position = start;
            EndPosition = end;
            Direction = end - start;
            Max_t = Direction.Length();
            Direction.Normalize();
        }
    }


    public class ITriangle
    {
        // fields
        public Vector3 P0;
        public Vector3 P1;
        public Vector3 P2;

        public Vector3 N;

        // constructor
        public ITriangle()
        {
            P0 = Vector3.Zero;
            P1 = Vector3.UnitX;
            P2 = Vector3.UnitZ;
        }

        public ITriangle(Vector3 p0, Vector3 p1, Vector3 p2)
        {
            P0 = p0;
            P1 = p1;
            P2 = p2;

            computeNormal();
        }

        // funcs
        public virtual void computeNormal()
        {
            Vector3 d1 = P1 - P0;
            Vector3 d2 = P2 - P0;

            N = Vector3.Cross(d1, d2);
            N.Normalize();
        }
    }

    public class ITexTriangle : ITriangle
    {
        // fields
        public Vector2 T0;
        public Vector2 T1;
        public Vector2 T2;

        // constructor
        public ITexTriangle(Vector3 p0, Vector3 p1, Vector3 p2)
            : base(p0, p1, p2)
        {
        }

        public virtual void setTexCoords(Vector2 t0, Vector2 t1, Vector2 t2)
        {
            T0 = t0;
            T1 = t1;
            T2 = t2;
        }
    }




    public static class GeometryHelper
    {

        #region operations on triangles

        /// <summary>
        /// Computes the distance between point and the triangle (v0, v1, v2). 
        /// Code taken from http://www.geometrictools.com/LibFoundation/Distance/Distance.html
        /// </summary>
        /// <param name="point"></param>
        /// <param name="v0"></param>
        /// <param name="v1"></param>
        /// <param name="v2"></param>
        /// <returns></returns>
        public static Vector3 closestPointInTriangle(Vector3 point, Vector3 v0, Vector3 v1, Vector3 v2)
        {
            {
                //  Vector3<Real> kDiff = m_rkTriangle.V[0] - m_rkVector;
                Vector3 kDiff = v0 - point;
                Vector3 kEdge0 = v1 - v0;
                Vector3 kEdge1 = v2 - v0;

                double fA00 = kEdge0.LengthSquared();

                double fA01 = Vector3.Dot(kEdge0, kEdge1);
                double fA11 = kEdge1.LengthSquared();
                double fB0 = Vector3.Dot(kDiff, kEdge0);
                double fB1 = Vector3.Dot(kDiff, kEdge1);
                double fC = kDiff.LengthSquared();
                double fDet = Math.Abs(fA00 * fA11 - fA01 * fA01);
                double fS = fA01 * fB1 - fA11 * fB0;
                double fT = fA01 * fB0 - fA00 * fB1;
                double fSqrDistance;

                if (fS + fT <= fDet)
                {
                    if (fS < 0.0)
                    {
                        if (fT < 0.0)  // region 4
                        {
                            if (fB0 < 0.0)
                            {
                                fT = 0.0;
                                if (-fB0 >= fA00)
                                {
                                    fS = 1.0;
                                    fSqrDistance = fA00 + 2.0 * fB0 + fC;
                                }
                                else
                                {
                                    fS = -fB0 / fA00;
                                    fSqrDistance = fB0 * fS + fC;
                                }
                            }
                            else
                            {
                                fS = 0.0;
                                if (fB1 >= 0.0)
                                {
                                    fT = 0.0;
                                    fSqrDistance = fC;
                                }
                                else if (-fB1 >= fA11)
                                {
                                    fT = 1.0;
                                    fSqrDistance = fA11 + 2.0 * fB1 + fC;
                                }
                                else
                                {
                                    fT = -fB1 / fA11;
                                    fSqrDistance = fB1 * fT + fC;
                                }
                            }
                        }
                        else  // region 3
                        {
                            fS = 0.0;
                            if (fB1 >= 0.0)
                            {
                                fT = 0.0;
                                fSqrDistance = fC;
                            }
                            else if (-fB1 >= fA11)
                            {
                                fT = 1.0;
                                fSqrDistance = fA11 + 2.0 * fB1 + fC;
                            }
                            else
                            {
                                fT = -fB1 / fA11;
                                fSqrDistance = fB1 * fT + fC;
                            }
                        }
                    }
                    else if (fT < 0.0)  // region 5
                    {
                        fT = 0.0;
                        if (fB0 >= 0.0)
                        {
                            fS = 0.0;
                            fSqrDistance = fC;
                        }
                        else if (-fB0 >= fA00)
                        {
                            fS = 1.0;
                            fSqrDistance = fA00 + 2.0 * fB0 + fC;
                        }
                        else
                        {
                            fS = -fB0 / fA00;
                            fSqrDistance = fB0 * fS + fC;
                        }
                    }
                    else  // region 0
                    {
                        // minimum at interior point
                        double fInvDet = 1.0 / fDet;
                        fS *= fInvDet;
                        fT *= fInvDet;
                        fSqrDistance = fS * (fA00 * fS + fA01 * fT + 2.0 * fB0) +
                            fT * (fA01 * fS + fA11 * fT + 2.0 * fB1) + fC;
                    }
                }
                else
                {
                    double fTmp0, fTmp1, fNumer, fDenom;

                    if (fS < 0.0)  // region 2
                    {
                        fTmp0 = fA01 + fB0;
                        fTmp1 = fA11 + fB1;
                        if (fTmp1 > fTmp0)
                        {
                            fNumer = fTmp1 - fTmp0;
                            fDenom = fA00 - 2.0f * fA01 + fA11;
                            if (fNumer >= fDenom)
                            {
                                fS = 1.0;
                                fT = 0.0;
                                fSqrDistance = fA00 + 2.0 * fB0 + fC;
                            }
                            else
                            {
                                fS = fNumer / fDenom;
                                fT = 1.0 - fS;
                                fSqrDistance = fS * (fA00 * fS + fA01 * fT + 2.0f * fB0) +
                                    fT * (fA01 * fS + fA11 * fT + 2.0) * fB1 + fC;
                            }
                        }
                        else
                        {
                            fS = 0.0;
                            if (fTmp1 <= 0.0)
                            {
                                fT = 1.0;
                                fSqrDistance = fA11 + 2.0 * fB1 + fC;
                            }
                            else if (fB1 >= 0.0)
                            {
                                fT = 0.0;
                                fSqrDistance = fC;
                            }
                            else
                            {
                                fT = -fB1 / fA11;
                                fSqrDistance = fB1 * fT + fC;
                            }
                        }
                    }
                    else if (fT < 0.0)  // region 6
                    {
                        fTmp0 = fA01 + fB1;
                        fTmp1 = fA00 + fB0;
                        if (fTmp1 > fTmp0)
                        {
                            fNumer = fTmp1 - fTmp0;
                            fDenom = fA00 - 2.0 * fA01 + fA11;
                            if (fNumer >= fDenom)
                            {
                                fT = 1.0;
                                fS = 0.0;
                                fSqrDistance = fA11 + 2.0 * fB1 + fC;
                            }
                            else
                            {
                                fT = fNumer / fDenom;
                                fS = 1.0 - fT;
                                fSqrDistance = fS * (fA00 * fS + fA01 * fT + 2.0 * fB0) +
                                    fT * (fA01 * fS + fA11 * fT + 2.0 * fB1) + fC;
                            }
                        }
                        else
                        {
                            fT = 0.0;
                            if (fTmp1 <= 0.0)
                            {
                                fS = 1.0;
                                fSqrDistance = fA00 + 2.0 * fB0 + fC;
                            }
                            else if (fB0 >= 0.0)
                            {
                                fS = 0.0;
                                fSqrDistance = fC;
                            }
                            else
                            {
                                fS = -fB0 / fA00;
                                fSqrDistance = fB0 * fS + fC;
                            }
                        }
                    }
                    else  // region 1
                    {
                        fNumer = fA11 + fB1 - fA01 - fB0;
                        if (fNumer <= 0.0)
                        {
                            fS = 0.0;
                            fT = 1.0;
                            fSqrDistance = fA11 + 2.0 * fB1 + fC;
                        }
                        else
                        {
                            fDenom = fA00 - 2.0f * fA01 + fA11;
                            if (fNumer >= fDenom)
                            {
                                fS = 1.0;
                                fT = 0.0;
                                fSqrDistance = fA00 + 2.0 * fB0 + fC;
                            }
                            else
                            {
                                fS = fNumer / fDenom;
                                fT = 1.0 - fS;
                                fSqrDistance = fS * (fA00 * fS + fA01 * fT + 2.0 * fB0) +
                                    fT * (fA01 * fS + fA11 * fT + 2.0 * fB1) + fC;
                            }
                        }
                    }
                }

                // account for numerical round-off error
                if (fSqrDistance < 0.0)
                {
                    fSqrDistance = 0.0;
                }

                //  m_kClosestPoint0 = m_rkVector;
                return (v0 + ((float)fS) * kEdge0 + ((float)fT) * kEdge1);
            }
        }

        /// <summary>
        /// Returns true if ptTest lies within the triangle defined by pt1, pt2, pt3
        /// Code taken from http://www.blackpawn.com/texts/pointinpoly/default.html
        /// </summary>
        /// <param name="f"></param>
        /// <param name="pt"></param>
        /// <returns></returns>
        public static bool pointInTriangle(Vector3 pt1, Vector3 pt2, Vector3 pt3, Vector3 ptTest)
        {
            // Compute vectors        
            Vector3 v0 = pt3 - pt1;
            Vector3 v1 = pt2 - pt1;
            Vector3 v2 = ptTest - pt1;

            // Compute dot products
            float dot00 = Vector3.Dot(v0, v0);
            float dot01 = Vector3.Dot(v0, v1);
            float dot02 = Vector3.Dot(v0, v2);
            float dot11 = Vector3.Dot(v1, v1);
            float dot12 = Vector3.Dot(v1, v2);

            // Compute barycentric coordinates
            float invDenom = 1 / (dot00 * dot11 - dot01 * dot01);
            float u = (dot11 * dot02 - dot01 * dot12) * invDenom;
            float v = (dot00 * dot12 - dot01 * dot02) * invDenom;

            if (u + v <= 1 && u > 0 && v > 0)
                return true;
            return false;
        }
        



        /// <summary>
        /// Intersects ray r with the triangle defined by (p1, p2, p3). Returns -1 if no intersection. Returns
        /// the intersection t if it intersects.
        /// </summary>
        /// <param name="r"></param>
        /// <param name="p1"></param>
        /// <param name="p2"></param>
        /// <param name="p3"></param>
        /// <returns></returns>
        public static float intersectRayTriangle(IRay r, Vector3 p1, Vector3 p2, Vector3 p3)
        {

            // computing barycentric coordinates of the triangle
            Vector3 e1 = p2 - p1;
            Vector3 e2 = p3 - p1;
            Vector3 s1 = Vector3.Cross(r.Direction, e2);
            float divisor = Vector3.Dot(s1, e1);
            if (divisor == 0)
                return -1;

            float invDivisor = 1.0f / divisor;

            // compute first baricentric coordinate b1
            Vector3 d = r.Position - p1;
            float b1 = Vector3.Dot(d, s1) * invDivisor;
            if ((b1 < 0.0f) || (b1 > 1.0f))
                return -1;

            // compute second baricentric coordinate b2
            Vector3 s2 = Vector3.Cross(d, e1);
            float b2 = Vector3.Dot(r.Direction, s2) * invDivisor;
            if ((b2 < 0.0f) || (b1 + b2 > 1.0f))
                return -1;

            // compute parameter t of the intersection
            float t = Vector3.Dot(e2, s2) * invDivisor;
            if ((t < 0) || (t > r.Max_t))
                return -1;

            // return intersection parameter t
            return t;

        }



        #endregion
        
        #region operations on xna standard primitives

        /// <summary>
        /// Computes the intersection point of ray and plane
        /// </summary>
        /// <param name="ray">the ray that intersects</param>
        /// <param name="plane">the plane that gets intersected</param>
        /// <returns></returns>
        public static Vector3 intersectRayPlane(Ray ray, Plane plane)
        {
            // check if paralell
            float dot = Vector3.Dot(plane.Normal, ray.Direction);
            if (Math.Abs(dot) <= 0.00001f)
            {
                return Vector3.Zero;
            }

            // compute intersection
            // Plane = P * N + d ==> d = - P * N|
            Vector3 P = Vector3.Multiply(plane.Normal, plane.D);
            Vector3 N = plane.Normal;
            float d = plane.D; // projection of (origin, circleCenter) to N

            Vector3 P0 = ray.Position;
            Vector3 V = ray.Direction;

            // compute parameter t on ray that intersects plane
            float t = -(Vector3.Dot(P0, N) + d) / (Vector3.Dot(V, N));

            // compute position on plane
            return P0 + Vector3.Multiply(V, t);
        }



        

        #endregion 

        # region operations on/with a BoundingSphere

        public static Vector3 getRandPos(BoundingSphere b, ref Random randomGen)
        {
            //Random randomGen = new Random();
            Vector3 p = new Vector3((float)randomGen.NextDouble(), (float)randomGen.NextDouble(), (float)randomGen.NextDouble());
            p = Vector3.Multiply(p, 2.0f);
            p = p - Vector3.One;
            while (p.Length() > 1.0f)
            {
                p = new Vector3((float)randomGen.NextDouble(), (float)randomGen.NextDouble(), (float)randomGen.NextDouble());
                p = Vector3.Multiply(p, 2.0f);
                p = p - Vector3.One;
            }

            return (b.Center + Vector3.Multiply(p, b.Radius));
        }





        #region intersect ccc functions


        public static bool vectorsShowToSameHalfSpace(Vector3 v1, Vector3 v2)
        {
            return Vector3.Dot(v1, v2) > 0.0f;
        }


        /// <summary>
        /// Computes whether a line segment intersects with a sphere
        /// </summary>
        /// <param name="start">segment's starting point</param>
        /// <param name="end">segment's ending poing</param>
        /// <param name="sphere">The sphere</param>
        /// <returns>
        /// Whether or not the line segment intersects with the sphere.
        /// </returns>
        public static bool intersectsWithLineSegment(Vector3 start, Vector3 end, BoundingSphere sphere)
        {
            bool intersects = false;
            // check if the points are inside the sphere
            if (sphere.Contains(start) == ContainmentType.Disjoint && sphere.Contains(end) == ContainmentType.Disjoint)
            {
                // see if the points are the same
                if ((start - end).Length() > 0.001f)
                {
                    // if the points are not inside, then check if the line intersects the sphere, and if the intersections point are acutally inside the sphere
                    IRay connection = new IRay(start, end);
                    Vector3 solution1, solution2;
                    if (GeometryHelper.getIntersectionOnSphere(connection, sphere, out solution1, out solution2) != ContainmentType.Disjoint)
                    {
                        // since _both_ intersection points lie on the line, the distances should sum up to the exact
                        // distance between the circle points, otherwise they are before or after the circle points
                        float closestLineLength = 2.0f * (start - end).Length();
                        float dist = (solution1 - start).Length() + (solution1 - end).Length();
                        dist += (solution2 - start).Length() + (solution2 - end).Length();
                        if (!(Math.Abs(closestLineLength - dist) > 0.001f))
                        {
                            intersects = true;
                        }
                    }
                }
            }
            else
            {
                intersects = true;
            }
            return intersects;
        }

        /// <summary>
        /// Computes the intersection points of a sphere and a ray.
        /// </summary>
        /// <param name="ray">ray</param>
        /// <param name="sphere">sphere</param>
        /// <param name="point1">first solution point</param>
        /// <param name="point2">second solution point</param>
        /// <returns>
        /// Returns a containment type which indicates if the ray hit, touched or
        /// intersected with the sphere. Returns as well the solution points if there was 
        /// an intersection or a touching. Intersection might be behind the ray origin.
        /// </returns>
        public static ContainmentType getIntersectionOnSphere(IRay ray, BoundingSphere sphere, out Vector3 point1, out Vector3 point2)
        {
            // s[1,2] = -(a/2) +/- sqrt(a^2/4 - b)
            // a = 2 * direction * (ray.origin - sphere.center)
            // b = |ray.origin - sphere.center|^2 - sphere.radius^2
            Vector3 temp = ray.Position - sphere.Center;
            Vector3 direction = ray.Direction;
            direction.Normalize();
            float a = 2.0f * Vector3.Dot(direction, temp);
            float b = Vector3.Dot(temp, temp) - sphere.Radius * sphere.Radius;

            float rootTerm = a * a / 4.0f - b;
            if (rootTerm < 0)
            {
                point1 = Vector3.Zero;
                point2 = point1;
                return ContainmentType.Disjoint;
            }
            else 
            {
                rootTerm = (float) Math.Sqrt(rootTerm);
                float solution1 = -a / 2.0f + rootTerm;
                float solution2 = -a / 2.0f - rootTerm;

                if (solution1 == solution2)
                {
                    // in this case the ray only touches the sphere -> one solution
                    //point1 = Vector3.Add(ray.Position, Vector3.Multiply(ray.Direction.Normalize, solution1));
                    point1 = ray.Position + solution1 * direction;
                    point2 = point1;
                    return ContainmentType.Contains;
                }
                else
                {
                    point1 = ray.Position + solution1 * direction;
                    point2 = ray.Position + solution2 * direction;
                    return ContainmentType.Intersects;
                }
            }
        }

        # endregion


        #endregion

        #region nearest point on IRay

        /// <summary>
        /// Computes the nearest position on the ray to the plane.
        /// </summary>
        /// <param name="ray">The Ray</param>
        /// <param name="planePosition">Position on the plane</param>
        /// <param name="planeNormal">Normal of the plane</param>
        /// <returns>Position on ray which is the nearest to the plane</returns>
        public static Vector3 getNearestPointOnRay(IRay ray, Vector3 planePosition, Vector3 planeNormal)
        {
            planeNormal.Normalize();

            // check if paralell
            if (Vector3.Dot(planeNormal, ray.Direction) == 0) return planePosition;

            // compute intersection
            // Plane = P * N + d ==> d = - P * N|
            Vector3 P = planePosition;
            Vector3 N = planeNormal;
            float d = -Vector3.Dot(P, N); // projection of (origin, circleCenter) to N

            Vector3 P0 = ray.Position;
            Vector3 V = ray.Direction;

            // compute parameter t on ray that intersects plane
            float t = -(Vector3.Dot(P0, N) + d) / (Vector3.Dot(V, N));

            // clamp t
            t = MathHelper.Clamp(t, 0, ray.Max_t);

            // compute position on plane
            return P0 + Vector3.Multiply(V, t);
        }

        public static Vector3 getNearestPointOnRayNoClamp(IRay ray, Vector3 planePosition, Vector3 planeNormal)
        {
            planeNormal.Normalize();

            // check if paralell
            if (Vector3.Dot(planeNormal, ray.Direction) == 0) return planePosition;

            // compute intersection
            // Plane = P * N + d ==> d = - P * N|
            Vector3 P = planePosition;
            Vector3 N = planeNormal;
            float d = -Vector3.Dot(P, N); // projection of (origin, circleCenter) to N

            Vector3 P0 = ray.Position;
            Vector3 V = ray.Direction;

            // compute parameter t on ray that intersects plane
            float t = -(Vector3.Dot(P0, N) + d) / (Vector3.Dot(V, N));

            // compute position on plane
            return P0 + Vector3.Multiply(V, t);
        }

        /// <summary>
        /// Returns the nearest position on targetRay to source ray
        /// </summary>
        /// <param name="targetRay">ray on which the position should be computed</param>
        /// <param name="sourceRay">ray to which the point is nearest</param>
        /// <returns>Position on targetRay that is nearest to sourceRay</returns>
        public static Vector3 getNearestPointOnRay(IRay targetRay, IRay sourceRay) {

            // compute normal to both ray directions
            Vector3 normal = Vector3.Cross(targetRay.Direction, sourceRay.Direction);
            
            // compute plane normal
            Vector3 plane_normal = Vector3.Cross(sourceRay.Direction, normal);

            // intersect targetRay with plane
            return getNearestPointOnRay(targetRay, sourceRay.Position, plane_normal);
        }

        public static Vector3 getNearestPointOnRayNoClamp(IRay targetRay, IRay sourceRay)
        {

            // compute normal to both ray directions
            Vector3 normal = Vector3.Cross(targetRay.Direction, sourceRay.Direction);

            // compute plane normal
            Vector3 plane_normal = Vector3.Cross(sourceRay.Direction, normal);

            // intersect targetRay with plane
            return getNearestPointOnRayNoClamp(targetRay, sourceRay.Position, plane_normal);
        }

        /// <summary>
        /// Returns the nearest position on targetRay to sourcePoint
        /// </summary>
        /// <param name="targetRay">ray on which the position should be computed</param>
        /// <param name="sourcePoint">position in 3D space</param>
        /// <returns>Position on ray (clamped with Max_t) which is closest to sourcePoint</returns>
        public static Vector3 getNearestPointOnRay(IRay targetRay, Vector3 sourcePoint)
        {
            // vector ray-point srcPoint
            Vector3 x = sourcePoint - targetRay.Position;

            // project x on ray
            float t = Vector3.Dot(targetRay.Direction, x);

            // clamp position on ray
            t = MathHelper.Clamp(t, 0.0f, targetRay.Max_t);

            // return result point
            return targetRay.Position + Vector3.Multiply(targetRay.Direction, t);
        }


        /// <summary>
        /// Returns the nearest position on targetRay to sourcePoint
        /// </summary>
        /// <param name="targetRay">ray on which the position should be computed</param>
        /// <param name="sourcePoint">position in 3D space</param>
        /// <returns>Position on ray which is closest to sourcePoint</returns>
        public static Vector3 getNearestPointOnRay(Ray targetRay, Vector3 sourcePoint)
        {
            // vector ray-point srcPoint
            Vector3 x = sourcePoint - targetRay.Position;

            // project x on ray
            float t = Vector3.Dot(targetRay.Direction, x);

            // return result point
            return targetRay.Position + Vector3.Multiply(targetRay.Direction, t);
        }

        #endregion

        #region additional functions for standard xna primitives

        public static Plane getPlane(Vector3 position, Vector3 normal) {
            Vector3 n = normal;
            n.Normalize();
            return new Plane(n, Vector3.Dot(n, Vector3.Zero - position));
        }

        public static Vector3 getClosestPointBetweenRays(IRay ray1, IRay ray2)
        {
            // use the direction of ray2 to build a plane with ray1 = tempPlane. Then use tempPlane's 
            // normal to build another plane tempPlane2 with ray2. This plane now intersects with ray1 
            // in point e. The two rays should be in the same plane, if they are not due to numerical 
            // errors we get a solution anyway. Since the difference between point e and f (second
            // intersection point on ray2) is not very big, it's not worth to calculate f.
            Plane tempPlane = getPlane(ray1.Position, Vector3.Cross(ray1.Direction, ray2.Direction));
            Plane tempPlane2 = getPlane(ray2.Position, Vector3.Cross(ray2.Direction, tempPlane.Normal));

            Vector3 e = intersectRayPlane(new Ray(ray1.Position, ray1.Direction), tempPlane2);
            return e;
        }

        public static Ray getRay(Vector3 start, Vector3 end) {
            Ray r = new Ray();
            r.Position = start;
            Vector3 d = end - start; d.Normalize();
            r.Direction = d;
            return r;
        }


        /// <summary>
        /// Calculates the point's distance from a specified plane.
        /// </summary>
        /// <param name="point">point, for which we want to know the distance</param>
        /// <param name="plane">plane, to which we want to know the distance</param>
        /// <param name="pointInPlane">Any point in the plane</param>
        /// <returns>
        /// The shortest distance of the point to the plane. If the result is zero, then the point
        /// is within the plane. If the result is negative, then the point is situated in the
        /// opposite half space than the normal vector is pointing to.
        /// </returns>
        public static float distancePointPlane(Vector3 point, Plane plane, Vector3 pointInPlane)
        {
            // hessesche normalform
            // determine missing coefficient d (ax + by + cz + d = 0) for formula:
            // n dot x + d = distance (a, b and c are already normalized components)
            plane.Normalize();
            float d = -Vector3.Dot(plane.Normal, pointInPlane);
            return Vector3.Dot(plane.Normal, point) + d;
        }

      
        #endregion

        public static float getAngle(Vector3 v1, Vector3 v2) { 
            // normalize vectors
            v1.Normalize();
            v2.Normalize();

            // compute cosine of angle
            float cosAlpha = Vector3.Dot(v1, v2);
            float sinAlpha = (Vector3.Cross(v1, v2)).Length();

            // return angle
            return (float)Math.Acos(cosAlpha);
        }


    }
}

